/*
* Copyright (C) 2007, CompHEP Collaboration
* Author: Alexander Sherstnev 
*------------------------------------------------
* $Id $
*
* $Log $
*
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "service2/include/drandXX.h"
#include "service2/include/syst2.h"
#include "out_ext.h"

#include "4_vector.h"
#include "alphas2.h"
#include "cut.h"
#include "kfcodes.h"
#include "kinaux.h"
#include "runVegas.h"
#include "strfun.h"
#include "subproc.h"
#include "vegas.h"
#include "evnt_tools.h"
#include "evnt_format_lhaef.h"

static FILE * events_;
static int * cChains;
static int nC;
static int cBasisPower;


static void write_event_lhaef (long cCub, int n, char * rand_state) {
  int i;
  int j;
  int icc;
  double f;
  int ntot_ = nout_ + nin_;
  double alphaem = 1/128.;
  int * cal = NULL;
  int * acal = NULL;

  cal = malloc (sizeof (int) * ntot_);
  acal = malloc (sizeof (int) * ntot_);
  for (i = 0; i < ntot_; i++) {
    cal[i] = 0;
    acal[i] = 0;
  }

  if (cBasisPower) {
    int cTag = 500;
    double sum = 0;
    for (i = 0; i < cBasisPower; i++) {
      sum += fabs (color_weights[i]);
    }
    sum *= drandXX ();
    for (i = 0; i < cBasisPower; i++) {
      sum -= fabs (color_weights[i]);
      if (sum <= 0) {
        break;
      }
    }
    if (i == cBasisPower) {
      i--;
    }
    if (color_weights[i] < 0) {
      n *= -1;
    }
    icc = i;

    for (j = 0; j < nC; j++) {
      cal[cChains[2 * nC * icc + 2 * j] - 1] = cTag;
      acal[cChains[2 * nC * icc + 2 * j + 1] - 1] = cTag;
      ++cTag;
    }
  }

  f = calcCutFactor ();
  if (!f) {
    fprintf (stderr, "Error : generator have written an event \n");
    fprintf (stderr, "        which does not pass cuts.\n");
    printDetailsOfCutFactor ();
  }

  for (i = 0; i < n; ++i) {
    double qcdscale = qcd_Scale_chep ();
    rnd_rotate_momentum ();

    fprintf (events_, "<event>\n");
    fprintf (events_, "%i 1 1.0 %17.10E %17.10E %17.10E\n", ntot_, qcdscale, alpha_2 (qcdscale), alphaem);
    for (j = 0; j < ntot_; ++j) {
      int id;
      int status = - 1;
      int mom1 = 0;
      int mom2 = 0;
      double pmass;
      vshortstr pname;
      pinf_ (proces_1.nsub, j + 1, pname, &pmass);
      id = kfpart (pname);
      if (j > 1) {
        status = 1;
        mom1 = 1;
        mom2 = 2;
      }
      fprintf (events_, "%i %i %i %i %i %i %17.10E %17.10E %17.10E %17.10E %17.10E 0.0 9.0\n", 
      id, status, mom1, mom2, cal[j], acal[j], 
      pvect[4 * j + 1], pvect[4 * j + 2], pvect[4 * j + 3], pvect[4 * j + 0], pmass);
    }
    fprintf (events_, "<event/>\n");
  }
  free(cal);
  free(acal);
}


static double cs = 0.0;
static double cserr = 0.0;

static void write_event_cap_lhef (void) {
  int i;
  int nprup;
  int idwtup;
  int pbeam[2];
  int pdfgup[2];
  int pdfsup[2];
  double ebeam[2];
  double sqrtS;
  Str_fun_Info strfun[2];

  fprintf (events_, "<LesHouchesEvents version=\"1.0\">\n");
  fprintf (events_, "<!-- File generated with CompHEP 4.5.0 -->\n");
  fprintf (events_, "<!-- \n"
                    "     Preliminary version, it is compatible with the Les Houches event file\n"
                    "     format (hep-ph/0609017), but contains no extra tags.\n"
                    "-->\n");
  fprintf (events_, "<header>\n");
  fprintf (events_, "<header/>\n");
  fprintf (events_, "<init>\n");
  vinf_ (0, NULL, &sqrtS);
  for (i = 0; i < 2; i++) {
    double p, e;
    wrt_sf_NF_ (i, &strfun[i]);

    e = (sqrtS * sqrtS + strfun[i].prt_mass * strfun[i].prt_mass - strfun[1 - i].prt_mass * strfun[1 - i].prt_mass) / (2 * sqrtS);
    p = sqrt (e * e - strfun[i].prt_mass * strfun[i].prt_mass);
    p = p * cosh (rapidity) + e * sinh (rapidity) * (1 - 2 * i);
    if (p * p < (10.E-10) * sqrtS) p = 0.0;
    ebeam[i] = p;

    pbeam[i] = kfbeam (strfun[i].prt_name);
    pdfgup[i] = strfun[i].PDFgr;
    pdfsup[i] = strfun[i].PDFid;
  }
  idwtup = 3;
  nprup = 1;

  fprintf (events_, "%i %i %17.10E %17.10E %i %i %i %i %i %i\n",
           pbeam[0], pbeam[1], ebeam[0], ebeam[1], pdfgup[0], pdfgup[1], pdfsup[0], pdfsup[1], idwtup, nprup);

  fprintf (events_, "%17.10E %17.10E %17.10E %i\n", cs, cserr, 1.0, 1);

  fprintf (events_, "<init/>\n");
}

static long fileEnd;

int prepare_evfile_lhaef (vegasGrid * vegPtr, double (*func) (double *, double), char * fname, 
   float * cubemaxval, int n_event, int n_cube, double max) {
  int status = 0;

  cStrings (proces_1.nsub, &nC, &cBasisPower, &cChains);
  if (cBasisPower) {
    color_weights = malloc (sizeof (double) * cBasisPower);
  }

  if (n_cube <= 0) {
   status = -2;
  }

  if (n_event <= 0) {
   status = -3;
  }

  if (max <= 0.0) {
   status = -4;
  }

  if (0 == status) {
    events_ = fopen (fname, "a+");
    fseek (events_, 0L, SEEK_END);
    if (0 == ftell (events_)) {
      write_event_cap_lhef ();
      status = vegas_events (vegPtr, n_cube, n_event, max, func, write_event_lhaef, cubemaxval);
    } else {
      int chck = CheckFormat (events_);
      if (4 != chck) {
        status = -1;
      } else {
        fseek (events_, 0L, SEEK_END);
        fileEnd = ftell (events_);
        status = vegas_events (vegPtr, n_cube, n_event, max, func, write_event_lhaef, cubemaxval);
      }
    }
  }
  fclose (events_);
  return status;
}

int complete_evfile_lhaef (char * fname, int store, int n_event, double mult, double rmax) {
  if (store) {
    integral.old = 1;
    if (integral.n_it) {
      double a, b;
      cs = integral.s1 / integral.s0;
      a = sqrt (integral.s0);
      b = fabs (integral.s1);
      cserr = a / b;
    }

    events_ = fopen (fname, "r+");
    write_event_cap_lhef ();
    fclose (events_);
  } else {
    truncate (fname, fileEnd);
  }
  return 1;
}
